/*
 * NSF Plugin For ProjectX
 * Copyright (C) 2002 Jason Cline
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include "resource.h"
#include "global.h"
#include "input.h"
#include "APU.h"
#include "VRC6.h"
#include "MMC5.h"
#include "m6502.h"

typedef struct tagNSFHEADER{
  UINT8  uFormat[5];
  UINT8  uVersion;
  UINT8  uTotalSongs;
  UINT8  uStartSong;
  UINT16 uLoadAddress;
  UINT16 uInitAddress;
  UINT16 uPlayAddress;
  INT8   szTitle[32];
  INT8   szArtist[32];
  INT8   szCopyright[32];
  UINT16 uNTSCSpeed;
  UINT8  uBanks[8];
  UINT16 uPALSpeed;
  UINT8  uSpeed;
  UINT8  uChips;
  UINT8  uExtra[4];
}NSFHEADER;

typedef struct tagCONFIG{
  INT8   nPan[20];
  UINT8  nOversample;
}CONFIG;

static NSFHEADER tHeader;
static CONFIG    tConfig;
static UINT32    uKillThread=0;
static HANDLE    hThread=INVALID_HANDLE_VALUE;
static UINT32    uFileLoaded=0;
static UINT8     uCurrentTrack=0;
static UINT8     *pCPU;
static UINT32    uCartridgeLength;
static UINT32    uBankswitchingEnabled=0;
static UINT8     *pCartridge=NULL;
static INT8      szLoadedFilename[255];
//required for 6502 emulation
UINT32           dwElapsedTicks=0;
UINT32           cyclesRemaining=0;
CONTEXTM6502     m6502;

//return the track number
UINT8 nsfTrackNumber(HWND hWnd);
//return the track name
INT8 *nsfTrackName(HWND hWnd);
//start everything
UINT32 nsfStartup(HWND hWnd);
//shutdown everything
UINT32 nsfShutdown(HWND hWnd);
//stop the play thread
UINT32 nsfStop(HWND hWnd);
//start the play thread up
UINT32 nsfPlay(HWND hWnd);
//go to prev track
UINT32 nsfPrev(HWND hWnd);
//go to next track
UINT32 nsfNext(HWND hWnd);
//load the file and start it playing
UINT32 nsfLoad(HWND hWnd, INT8 * szFilename);
//handles writes to the cpu
void nsfWrite(UINT32 uAddress, UINT8 uValue, struct MemoryWriteByte *pMem);
//handles reads from the cpu
UINT8 nsfRead(UINT32 uAddress, struct MemoryReadByte *pMem);
//config the plug-in
UINT32 nsfConfig(HWND hWnd);
//learn about the plug-in
UINT32 nsfAbout(HWND hWnd);
//get the song info
UINT32 nsfTrackInfo(HWND hWnd);

static INPUTPLUGIN inputNSF={
  nsfLoad,
  nsfPlay,
  nsfStop,
  nsfPrev,
  nsfNext,
  nsfTrackNumber,
  nsfTrackName,
  NULL, //nsf songs don't have a length
  nsfTrackInfo,
  nsfConfig,
  nsfAbout,
  nsfStartup,
  nsfShutdown,
  "Nintendo Sound File=*.nsf;",
  //the rest is filled in by the calling program
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};

struct MemoryWriteByte tNSFWrite[]=
{
  {0x0000, 0xffff, nsfWrite},
  {(UINT32) -1, (UINT32) -1, NULL}
};

struct MemoryReadByte tNSFRead[] =
{
  {0x0000, 0xffff, nsfRead},
  {(UINT32) -1, (UINT32) -1, NULL}
};

//print a formatted message to the screen
void Message(char *string, ...)
{
va_list arg;
char szBuffer[1024];

  va_start(arg, string);
  vsprintf(szBuffer, string, arg);
  va_end(arg);
  MessageBox(NULL, szBuffer, "Meridian Prime NSF Plug-in", MB_OK);
}

//perform bankswitching
void nsfBankswitch(UINT32 uAddress, UINT8 uValue)
{
UINT32 uOffset=0;
UINT32 uLength=0x1000;

  //exit if this cartridge isnt bankswitched
  if(uBankswitchingEnabled==0) return;
  //otherwise do the nasty
  uOffset=0x8000 + (tHeader.uLoadAddress & 0x0fff) + ((uAddress & 0x07) * 0x1000);
  if(0x10000 - uOffset < 0x1000) uLength=0x10000 - uOffset;
  CopyMemory(pCPU + uOffset, pCartridge + (uValue * 0x1000), uLength);
}

//handles writes to the cpu
void nsfWrite(UINT32 uAddress, UINT8 uValue, struct MemoryWriteByte *pMem)
{
  //apu writes
  if(uAddress>=0x4000 && uAddress<=0x4015) mAPUWrite(uAddress, uValue);
  //bankswitching
  else if(uAddress>=0x5ff8 && uAddress<=0x5fff) nsfBankswitch(uAddress, uValue);
  //vrc6
  else if(uAddress>=0x9000 && uAddress<=0x9002 && (tHeader.uChips & 0x1)) mVRC6Write(uAddress, uValue);
  else if(uAddress>=0xA000 && uAddress<=0xA002 && (tHeader.uChips & 0x1)) mVRC6Write(uAddress, uValue);
  else if(uAddress>=0xB000 && uAddress<=0xB002 && (tHeader.uChips & 0x1)) mVRC6Write(uAddress, uValue);
  //mmc5
  else if(((uAddress>=0x5000 && uAddress<=0x5015) || (uAddress>=0x5205 && uAddress<=0x5206)) && (tHeader.uChips & 0x8)) mMMC5Write(uAddress, uValue);
  //set the byte in memory
  else pCPU[uAddress]=uValue;
}

//handles reads from the cpu
UINT8 nsfRead(UINT32 uAddress, struct MemoryReadByte *pMem)
{
  if(uAddress==0x4015) return(mAPURead(uAddress));
  else if(uAddress>=0x5205 && uAddress<=0x5206  && (tHeader.uChips & 0x8)) return(mMMC5Read(uAddress));
  else return(pCPU[uAddress]);
}

//play the music!
DWORD WINAPI nsfPlayThread(void *uStop)
{
UINT8  uBuffer[2940];
UINT32 i, j;
INT32  nLeft, nRight;
INT32  nSquareA, nSquareB, nTriangle, nNoise, nDPCM, nVRC6SquareA, nVRC6SquareB, nVRC6Sawtooth, nMMC5SquareA, nMMC5SquareB, nMMC5DPCM;

  //decode the music
  while(!*((INT32 *)uStop)){
    //can we write to the sound buffer?
    if(inputNSF.audioCanWrite(2940)){
      //play the song for 1 frame
      pCPU[0x5ff8]=0x20;
      pCPU[0x5ff9]=tHeader.uPlayAddress & 0xff;
      pCPU[0x5ffa]=(tHeader.uPlayAddress >> 8) & 0xff;
      pCPU[0x5ffb]=0xf2;
      m6502GetContext(&m6502);
      m6502.m6502pc=0x5ff8;
      m6502SetContext(&m6502);
      m6502exec(1790000 / 60);
      //now emulate the sound
      for(i=0; i<735; i++){
        nSquareA=nSquareB=nTriangle=nNoise=nDPCM=nVRC6SquareA=nVRC6SquareB=nVRC6Sawtooth=nMMC5SquareA=nMMC5SquareB=nMMC5DPCM=0;
        for(j=0; j<tConfig.nOversample; j++){
          //process apu
          nSquareA+=mAPUSquareUpdateA();
          nSquareB+=mAPUSquareUpdateB();
          nTriangle+=mAPUTriangleUpdate();
          nNoise+=mAPUNoiseUpdate();
          nDPCM+=mAPUDPCMUpdate();
          //process vrc6
          if(tHeader.uChips & 0x1){
            nVRC6SquareA+=mVRC6SquareUpdateA();
            nVRC6SquareB+=mVRC6SquareUpdateB();
            nVRC6Sawtooth+=mVRC6SawtoothUpdate();
          }
          //process mmc5
          if(tHeader.uChips & 0x8){
            nMMC5SquareA+=mMMC5SquareUpdateA();
            nMMC5SquareB+=mMMC5SquareUpdateB();
            nMMC5DPCM+=mMMC5PCMUpdate();
          }
        }
        nLeft=nSquareA + nTriangle + nNoise + nDPCM + nVRC6SquareA + nVRC6Sawtooth + nMMC5SquareA + nMMC5DPCM;
        nRight=nSquareB + nTriangle + nNoise + nDPCM + nVRC6SquareB + nVRC6Sawtooth + nMMC5SquareA + nMMC5DPCM;
        nLeft/=tConfig.nOversample;
        nRight/=tConfig.nOversample;
        uBuffer[(i*4)+0]=(nLeft >> 8) & 0xff;
        uBuffer[(i*4)+1]=nLeft & 0xff;
        uBuffer[(i*4)+2]=(nRight >> 8) & 0xff;
        uBuffer[(i*4)+3]=nRight & 0xff;
      }
      inputNSF.audioWrite(uBuffer, 2940);
    }
    //if not sleep for 20 milliseconds
    else Sleep(20);
  }
  return(0);
}

//return the track number
UINT8 nsfTrackNumber(HWND hWnd)
{
  //if no file is loaded then exit
  if(uFileLoaded==0) return(0);  
  else return(uCurrentTrack);
}

//return the track name
INT8 *nsfTrackName(HWND hWnd)
{
  //if no file is loaded then exit
  if(uFileLoaded==0) return(NULL);  
  else return(tHeader.szTitle);
}

//start everything
UINT32 nsfStartup(HWND hWnd)
{
INT8 *szData;

  //clear config
  ZeroMemory(&tConfig, sizeof(tConfig));
  //set up default values in case there is no ini file
  tConfig.nOversample=1;
  szData=inputNSF.iniRead("NSF", "Oversample"); sscanf(szData, "%i", &tConfig.nOversample);
  szData=inputNSF.iniRead("NSF", "Panning"); sscanf(szData, "%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i", &tConfig.nPan[0], &tConfig.nPan[1], &tConfig.nPan[2], &tConfig.nPan[3], &tConfig.nPan[4], &tConfig.nPan[5], &tConfig.nPan[6], &tConfig.nPan[7], &tConfig.nPan[8], &tConfig.nPan[9], &tConfig.nPan[10]);
  return(1);
}

//shutdown everything
UINT32 nsfShutdown(HWND hWnd)
{
  //free memory
  if(pCartridge) free(pCartridge); pCartridge=NULL;
  if(pCPU) free(pCPU); pCPU=NULL;
  //write the config file
  inputNSF.iniWrite("NSF", "Oversample", "%i", tConfig.nOversample);
  inputNSF.iniWrite("NSF", "Panning", "%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i", tConfig.nPan[0], tConfig.nPan[1], tConfig.nPan[2], tConfig.nPan[3], tConfig.nPan[4], tConfig.nPan[5], tConfig.nPan[6], tConfig.nPan[7], tConfig.nPan[8], tConfig.nPan[9], tConfig.nPan[10]);
  return(1);
}

//stop the play thread
UINT32 nsfStop(HWND hWnd)
{
  //stop the thread
  if(hThread!=INVALID_HANDLE_VALUE){
    uKillThread=1;
//    if(WaitForSingleObject(hThread, INFINITE)==WAIT_TIMEOUT){
      TerminateThread(hThread, 0);
//    }
    CloseHandle(hThread);
    hThread=INVALID_HANDLE_VALUE;
  }
  //stop the sound engine
  inputNSF.audioShutdown(hWnd);
  return(0);
}

//start the play thread up
UINT32 nsfPlay(HWND hWnd)
{
INT32 uThread;

  //if no file is loaded then exit
  if(uFileLoaded==0) return(0);
  //first stop anything that is playing
  nsfStop(hWnd);
  //clear the cpu
  ZeroMemory(pCPU, 0x10000);
  if(uBankswitchingEnabled==0){
    CopyMemory(pCPU + tHeader.uLoadAddress, pCartridge, uCartridgeLength);
  }
  else{
//    Message("Load Address: 0x%04x (0x%03x)\nInit Address: 0x%4x\nPlay Address: 0x%04x\nBanks: %i %i %i %i %i %i %i %i", tHeader.uLoadAddress, (tHeader.uLoadAddress & 0x0fff), tHeader.uInitAddress, tHeader.uPlayAddress, tHeader.uBanks[0], tHeader.uBanks[1], tHeader.uBanks[2], tHeader.uBanks[3], tHeader.uBanks[4], tHeader.uBanks[5], tHeader.uBanks[6], tHeader.uBanks[7]);
    //otherwise do the bankswitching if needed
    nsfBankswitch(0x5ff8, tHeader.uBanks[0]);
    nsfBankswitch(0x5ff9, tHeader.uBanks[1]);
    nsfBankswitch(0x5ffa, tHeader.uBanks[2]);
    nsfBankswitch(0x5ffb, tHeader.uBanks[3]);
    nsfBankswitch(0x5ffc, tHeader.uBanks[4]);
    nsfBankswitch(0x5ffd, tHeader.uBanks[5]);
    nsfBankswitch(0x5ffe, tHeader.uBanks[6]);
    nsfBankswitch(0x5fff, tHeader.uBanks[7]);
  }
  //reset the sound hardware
  mAPUReset(44100 * tConfig.nOversample, pCPU);
  if(tHeader.uChips & 0x1) mVRC6Reset(44100 * tConfig.nOversample);
  if(tHeader.uChips & 0x8) mMMC5Reset(44100 * tConfig.nOversample);

  //init the song
  pCPU[0x5ff8]=0x20;
  pCPU[0x5ff9]=tHeader.uInitAddress & 0xff;
  pCPU[0x5ffa]=(tHeader.uInitAddress >> 8) & 0xff;
  pCPU[0x5ffb]=0xf2;
  m6502GetContext(&m6502);
  m6502.m6502pc=0x5ff8;
  m6502.m6502af=(uCurrentTrack-1);
  m6502.m6502x=0x00;
  m6502SetContext(&m6502);
  m6502exec(1790000/60);
  //start the sound engine
  inputNSF.audioStartup(hWnd, 44100, 16);
  //now start up the music
  uKillThread=0;
  hThread=(HANDLE)CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)nsfPlayThread, (void *)&uKillThread, 0, &uThread);
//  SetThreadPriority(hThread, THREAD_PRIORITY_NORMAL);
  return(1);
}

//go to prev track
UINT32 nsfPrev(HWND hWnd)
{
  //if no file is loaded then exit
  if(uFileLoaded==0) return(0);
  //first stop anything that is playing
  nsfStop(hWnd);
  //prev track
  uCurrentTrack--;
  if(uCurrentTrack==0) uCurrentTrack=tHeader.uTotalSongs;
  //now start up the music
  nsfPlay(hWnd);
  return(1);
}

//go to next track
UINT32 nsfNext(HWND hWnd)
{
  //if no file is loaded then exit
  if(uFileLoaded==0) return(0);
  //first stop anything that is playing
  nsfStop(hWnd);
  //next track
  uCurrentTrack++;
  if(uCurrentTrack>tHeader.uTotalSongs) uCurrentTrack=1;
  //now start up the music
  nsfPlay(hWnd);
  return(1);
}

//swap the bytes of a 16bit value
UINT16 nsfByteSwap(UINT16 uValue)
{
UINT8  uHi, uLo;

  uHi=uValue >> 8;
  uLo=uValue & 0xff;
  return((uLo << 8) | uHi);
}

//load the file and start it playing
UINT32 nsfLoad(HWND hWnd, INT8 * szFilename)
{
FILE *fp=NULL;
UINT16 uTmp=0;

  //first stop anything that is playing
  nsfStop(hWnd);
  //if a file is loaded then free it
  if(uFileLoaded!=0){
    if(pCartridge) free(pCartridge); pCartridge=NULL;
    if(pCPU) free(pCPU); pCPU=NULL;
    uFileLoaded=0;
  }
  //load the file
  fp=fopen(szFilename, "rb");
  if(fp==NULL) return(0);
  fseek(fp, 0, SEEK_END);
  uCartridgeLength=ftell(fp)-0x80;
  fseek(fp, 0, SEEK_SET);
  fread(&tHeader, 1, 0x80, fp);
  pCartridge=(UINT8 *)malloc(uCartridgeLength);
  fread(pCartridge, 1, uCartridgeLength, fp);
  fclose(fp);
  //set some initial values
  uCurrentTrack=tHeader.uStartSong;
  uBankswitchingEnabled=tHeader.uBanks[0] | tHeader.uBanks[1] | tHeader.uBanks[2] | tHeader.uBanks[3] | tHeader.uBanks[4] | tHeader.uBanks[5] | tHeader.uBanks[6] | tHeader.uBanks[7];
  //file is loaded
  uFileLoaded=1;
  //create the cpu
  pCPU=(UINT8 *)malloc(0x10000);
  m6502.m6502Base=pCPU;
  m6502.m6502MemoryRead=tNSFRead;
  m6502.m6502MemoryWrite=tNSFWrite;
  m6502SetContext(&m6502);
  //copy the filename for later use
  strcpy(szLoadedFilename, szFilename);
  //now start up the music
  nsfPlay(hWnd);
  return(1);
}

//handles the track info window messages
BOOL CALLBACK TrackInfoDialogWindowProc(HWND hWnd, UINT32 uMsg, WPARAM wParam, LPARAM lParam)
{
INT8 szTitle[32];
INT8 szArtist[32];
INT8 szCopyright[32];
HWND hEditWnd;
FILE *fp;
INT8 szInfo[1024];

/*
#define NSF_SONG_INFO_LOCATION          1000
#define NSF_TRACK_INFO_DETAILS          1001
#define NSF_SONG_INFO_TITLE             1002
#define NSF_SONG_INFO_ARTIST            1003
#define NSF_SONG_INFO_COPYRIGHT         1004
#define NSF_SONG_INFO_UPDATE            1005
#define NSF_SONG_INFO_CLOSE             1006
*/
  switch(uMsg){
    case WM_INITDIALOG:
      //show the file location
      SetDlgItemText(hWnd, NSF_TRACK_INFO_LOCATION, szLoadedFilename);
      //set the limit for each edit control to 31 characters
      SendDlgItemMessage(hWnd, NSF_TRACK_INFO_TITLE, EM_SETLIMITTEXT, 31, 0);
      SendDlgItemMessage(hWnd, NSF_TRACK_INFO_ARTIST, EM_SETLIMITTEXT, 31, 0);
      SendDlgItemMessage(hWnd, NSF_TRACK_INFO_COPYRIGHT, EM_SETLIMITTEXT, 31, 0);
      //copy the track info into the edit controls
      SetDlgItemText(hWnd, NSF_TRACK_INFO_TITLE, tHeader.szTitle);
      SetDlgItemText(hWnd, NSF_TRACK_INFO_ARTIST, tHeader.szArtist);
      SetDlgItemText(hWnd, NSF_TRACK_INFO_COPYRIGHT, tHeader.szCopyright);
      //fill in the track info
      sprintf(szInfo, "Size: %i bytes\r\nTracks: %i\r\nBankswitched: %s\r\nChips: %s%s%s%s%s%s%s", uCartridgeLength + 0x80, tHeader.uTotalSongs, uBankswitchingEnabled ? "Yes" : "No", "APU", (tHeader.uChips & 0x01) ? ", VRC6" : "", (tHeader.uChips & 0x02) ? ", VRC7" : "", (tHeader.uChips & 0x04) ? ", FDS" : "", (tHeader.uChips & 0x08) ? ", MMC5" : "", (tHeader.uChips & 0x10) ? ", N-106" : "", (tHeader.uChips & 0x20) ? ", FME-07" : "");
      SetDlgItemText(hWnd, NSF_TRACK_INFO_DETAILS, szInfo);
      return(1);
    case WM_COMMAND:
      switch(wParam){
        //update the current info and the file
        case NSF_TRACK_INFO_UPDATE:
          ZeroMemory(szTitle, sizeof(szTitle));
          ZeroMemory(szArtist, sizeof(szArtist));
          ZeroMemory(szCopyright, sizeof(szCopyright));
          hEditWnd=GetDlgItem(hWnd, NSF_TRACK_INFO_TITLE);
          GetWindowText(hEditWnd, szTitle, 31);
          hEditWnd=GetDlgItem(hWnd, NSF_TRACK_INFO_ARTIST);
          GetWindowText(hEditWnd, szArtist, 31);
          hEditWnd=GetDlgItem(hWnd, NSF_TRACK_INFO_COPYRIGHT);
          GetWindowText(hEditWnd, szCopyright, 31);
          strcpy(tHeader.szTitle, szTitle);
          strcpy(tHeader.szArtist, szArtist);
          strcpy(tHeader.szCopyright, szCopyright);
          if((fp=fopen(szLoadedFilename, "r+b"))!=NULL){
            fseek(fp, 0x0e, SEEK_SET);
            fwrite(szTitle, 1, 32, fp);
            fwrite(szArtist, 1, 32, fp);
            fwrite(szCopyright, 1, 32, fp);
            fclose(fp);
          }
          EndDialog(hWnd, 0);
          return(1);
        //close the dialog box without updating
        case NSF_TRACK_INFO_CLOSE:
          EndDialog(hWnd, 0);
          return(1);
      }
      return(1);
    case WM_CLOSE:
      EndDialog(hWnd, 0);
      return(1);
  }
  return(0);
}

//get the track info
UINT32 nsfTrackInfo(HWND hWnd)
{
  DialogBox(inputNSF.hInstance, MAKEINTRESOURCE(NSF_TRACK_INFORMATION_DIALOG), hWnd, (DLGPROC)TrackInfoDialogWindowProc);
  return(1);
}

//config the plug-in
UINT32 nsfConfig(HWND hWnd)
{
  MessageBox(hWnd, "No Config Options Available At This Time", "Meridian Prime NSF Configuration", MB_OK);
  return(1);
}

//learn about the plug-in
UINT32 nsfAbout(HWND hWnd)
{
  MessageBox(hWnd, "Meridian Prime NSF Player v0.0\nCopyright (c) 2003 Jason Cline", "Meridian Prime NSF Player", MB_OK);
  return(1);
}

//export function that sends the plugin information back to the main program
__declspec( dllexport ) INPUTPLUGIN *exportMeridianPrimeGetInModule()
{
  return &inputNSF;
}