//input plugin interface structure
typedef struct tagINPUTPLUGIN{
  //called when the load/eject button is hit
  UINT32    (*processLoad)(HWND, INT8 *);
  //called when the play button is hit
  UINT32    (*processPlay)(HWND);
  //called when the stop button is hit
  UINT32    (*processStop)(HWND);
  //called when the prev button is hit
  UINT32    (*processPrev)(HWND);
  //called when the next button is hit
  UINT32    (*processNext)(HWND);
  //this should return the track number playing
  UINT8     (*getTrackNumber)(HWND hWnd);
  //this should return the name of the song playing
  INT8      *(*getTrackName)(HWND hWnd);
  //this should return the play length of the track in seconds
  UINT32    (*getTrackLength)(HWND, UINT32 **nLength);
  //called when someone clicks on the song title
  UINT32    (*processSongInfo)(HWND hWnd);
  //called when someone wants to config the plugin
  UINT32    (*processConfig)(HWND hWnd);
  //called when someone wants to learn about the plugin
  UINT32    (*processAbout)(HWND hWnd);
  //called once when the plugin is loaded
  UINT32    (*inputStartup)(HWND hWnd);
  //called once when the plugin is unloaded
  UINT32    (*inputShutdown)(HWND hWnd);
  //what files this plugin supports, in the following format examples:
  //Nintendo Sound File=*.nsf
  //Super Nintendo Sound File=*.spc;*.s01;*.s02
  INT8      *szFilter;
  //=========================================================================
  //everything below this is filled in by the loading program and not the dll
  //=========================================================================
  //this is the intance of the dll in case you need to load a resource from within it
  HINSTANCE hInstance;
  //call this to shutdown all audio engine
  void      (*audioShutdown)();
  //call this to startup the audio engine
  UINT32    (*audioStartup)(HWND hWnd, UINT32 nSamplerate, UINT16 nBPS);
  //call this to find out if you can write to the buffer
  UINT32    (*audioCanWrite)(UINT32 nSize);
  //call this to write to the buffer
  void      (*audioWrite)(INT8 *pBuffer, UINT32 nSize);
  //read a key and its values from an ini file
  INT8     *(*iniRead)(INT8 *szSection, INT8 *szKey);
  //write a key and its values to an ini file
  void      (*iniWrite)(INT8 *szSection, INT8 *szKey, INT8 *szData, ...);
}INPUTPLUGIN;

//build the filter for the open file dialog
INT8 *getInputFilter();
//return a pointer to an input dll based on the filename
INPUTPLUGIN *getInputDLL(INT8 *szFilename);
//free the loaded dll's
UINT32 freeInputDLLs();
//load an input dll after it has been found
UINT32 loadInputDLLs(HWND, INT8 *);
//scan for input dll's and then call the load routine
UINT32 scanInputDLLs(HWND, INT8 *);
//return the plugin that corresponds to the file extension passed
INPUTPLUGIN *getPlugin(INT8 *);